import json
import pickle
import random
import sentencepiece as spm
import math
import numpy as np
import Utils
import multiprocessing as mp
import os

class SG_BPE_Models:
    def __init__(self, experiment_name, is_continue_execution, final_vocab_size, initial_vocab_size, partial_corpus_filepath):
        # SG-BPE model - this we are gonna update
        sg_bpe_model_prefix = "results/{}/sg_bpe".format(experiment_name)
        sg_bpe_model = sg_bpe_model_prefix + ".model"

        if not is_continue_execution:
            spm.SentencePieceTrainer.train(input=partial_corpus_filepath, model_prefix=sg_bpe_model_prefix, vocab_size=initial_vocab_size, model_type="bpe")
        self._sg_bpe_model = spm.SentencePieceProcessor(model_file=sg_bpe_model)

        # Vanilla BPE model - for comparisons
        vanilla_bpe_model_prefix = "results/{}/bpe_vanilla".format(experiment_name)
        vanilla_bpe_model = vanilla_bpe_model_prefix + ".model"
        
        if not is_continue_execution:
            spm.SentencePieceTrainer.train(input=partial_corpus_filepath, model_prefix=vanilla_bpe_model_prefix, vocab_size=final_vocab_size, model_type="bpe")
        self._bpe_vanilla_model = spm.SentencePieceProcessor(model_file=vanilla_bpe_model)

    def get_sg_bpe_model(self):
        return self._sg_bpe_model

    def get_bpe_vanilla_model(self):
        return self._bpe_vanilla_model


class Model:
    def __init__(self, experiment_name, log, model, model_name, \
        target_embeddings, context_embeddings, corpus_lines, max_lines_per_token, window_size, is_continue_execution=False, vocab_filepath=False):

        self._experiment_name = experiment_name
        self._model = model
        self._model_name = model_name
        self._target_embeddings = target_embeddings
        self._context_embeddings = context_embeddings
        self._log = log
        self._current_vocab = None
        self._should_compute_vocab = True
        self._corpus_lines = corpus_lines
        self._max_lines_per_token = max_lines_per_token
        self._window_size = window_size
              
        if is_continue_execution:
            with open(vocab_filepath + ".bin", "rb") as vocab_file:
                current_bpe_vocab = pickle.load(vocab_file)
                self.set_vocab(current_bpe_vocab)

    def initialize_encoded_form_for_corpus_lines(self):
        model_encoded_coprus_lines_token_ids = []
        model_encoded_coprus_lines_token_pieces = []
        for line in self._corpus_lines:
            tokens_in_line_ints = self._model.encode(line, out_type=int)
            tokens_in_line_pieces = [self._model.id_to_piece(x) for x in tokens_in_line_ints]

            model_encoded_coprus_lines_token_ids.append(tokens_in_line_ints)
            model_encoded_coprus_lines_token_pieces.append(tokens_in_line_pieces)

        self._model_encoded_corpus_lines_token_ids = model_encoded_coprus_lines_token_ids
        self._model_encoded_corpus_lines_token_pieces = model_encoded_coprus_lines_token_pieces

    def initialize_token_to_line_indices_dictionary(self, current_vocab, corpus_lines, experiment_name, is_continue_execution):
        token_to_line_indices_dict_filepath = "results/{}/token_to_line_indices_dict.bin".format(experiment_name)
        if is_continue_execution and os.path.exists(token_to_line_indices_dict_filepath):
            with open(token_to_line_indices_dict_filepath, "rb") as token_to_line_indices_dict_file:
                self._token_to_line_indices_dict = pickle.load(token_to_line_indices_dict_file)
        else:
            self._token_to_line_indices_dict = Utils.token_to_line_indices_dictionary(current_vocab, corpus_lines)
            with open(token_to_line_indices_dict_filepath, "wb") as token_to_line_indices_dict_file:
                pickle.dump(self._token_to_line_indices_dict, token_to_line_indices_dict_file)
        
    def update_encoded_form_for_corpus_lines(self, tokens_pruned):
        for token in tokens_pruned:
            line_indices_with_token = self._token_to_line_indices_dict[token]
            for index in line_indices_with_token:
                current_line = self._corpus_lines[index]
                self._model_encoded_corpus_lines_token_ids[index] = self._model.encode(current_line, out_type=int)
                self._model_encoded_corpus_lines_token_pieces[index] = [self._model.id_to_piece(x) for x in self._model_encoded_corpus_lines_token_ids[index]]

	# We use the "get_current_vocab" method because of sentencepiece tricky way to remove tokens from the vocabulary.
    def get_current_vocab(self):
        if not self._should_compute_vocab:
            return self._current_vocab

        model_vocab = []
        for i in range(self._model.vocab_size()):
            model_vocab.append(self._model.id_to_piece(i))

        return model_vocab

    def log_experiments_model_results(self, training_filepath, model_name_override=None):
        if model_name_override:
            model_name = model_name_override
        else:
            model_name = self._model_name
        
        ## Log model vocabulary
        model_vocab = self.get_current_vocab()
        model_vocab_filepath = ("./results/{}/" + model_name + "_vocab.txt").format(self._experiment_name)
        with open(model_vocab_filepath, "w+") as model_vocab_results_file:
            model_vocab_results_file.write(json.dumps(model_vocab, indent=4))

        ## Log encoding of input file with model vocabulary
        with open(training_filepath) as input_file:
            input_file_data = input_file.read()

        encoded_data = [self._model.id_to_piece(x) for x in self._model.encode(input_file_data)]
        model_encoding_filepath = ("./results/{}/" + model_name + "_encoding.txt").format(self._experiment_name)
        with open(model_encoding_filepath, "w+") as encoding_results_file:
            encoding_results_file.write(' '.join(encoded_data))

    def sg_for_window(self, target_token, window):
        current_p = 0
        for w in window:
            # calculate current value to add
            dot_product = np.dot(self._target_embeddings[target_token], self._context_embeddings[w])
            try:
                current_p += math.log(Utils.sigmoid(dot_product))
            except:
                pass

        return (-1) * current_p

    def token_context_sg_log_prob(self, token_int, i, tokens_in_line_ints):
        # get context window tokens
        window, _, _ = Utils.compute_window(i, tokens_in_line_ints, self._window_size)
        return self.sg_for_window(token_int, window)

    def total_sg_log_prob(self, training_filepath):
        # getting lines from corpus
        with open(training_filepath, "r") as training_file:
            corpus_lines = training_file.readlines()

        p = 0
        for line in corpus_lines:
            current_p = 0
            tokens_in_line_ints = self._model.encode(line, out_type=int)
            for i, token_int in enumerate(tokens_in_line_ints):
                current_p += self.token_context_sg_log_prob(token_int, i, tokens_in_line_ints)

            p += current_p

        return p

    # Compute total_sg_log_prob without each token
    # -- This is executed once in iteration --
    # Multi Processing version of this method ###########
    def get_sg_log_prob_without_tokens_mp(self, current_total_sg, training_filepath, nat_list=None, dict_of_top_tokens=None):
        token_and_sg_log_prob_without_it = {}
        current_vocab = self.get_current_vocab()

        if dict_of_top_tokens:
            current_vocab = dict_of_top_tokens

        print("\nCurrent vocab len - {}".format(len(current_vocab)))

        if nat_list:
            current_nat_list = Utils.get_not_ablateable_tokens_list(current_vocab)
            current_vocab = [t for t in current_vocab if t not in current_nat_list]

        process_pool = mp.Pool(mp.cpu_count())
        params = [(self._model, token, current_total_sg, \
                    current_vocab, training_filepath, \
                    self._target_embeddings, self._context_embeddings, self._log, self._corpus_lines, self._window_size) for token in current_vocab]
        res = process_pool.starmap(Utils.sg_wo_token_mp, params)
        
        for r in res:
            token = r[0]
            current_sg_log_prob = r[1]
            token_and_sg_log_prob_without_it[token] = current_sg_log_prob

        return token_and_sg_log_prob_without_it

    # More optimized form of "get_sg_log_prob_without_tokens_mp"
    def get_sg_log_prob_without_tokens_mp2(self, current_total_sg, nat_list=None, dict_of_top_tokens=None):
        token_and_sg_log_prob_without_it = {}
        
        current_vocab = self.get_current_vocab()
        if dict_of_top_tokens:
            current_vocab = dict_of_top_tokens

        if nat_list:
            current_nat_list = Utils.get_not_ablateable_tokens_list(current_vocab)
            current_vocab = [t for t in current_vocab if t not in current_nat_list]

        process_pool = mp.Pool(mp.cpu_count())
        lines_per_token = {}

        params = []
        for token in current_vocab:
            lines_to_consider = self._token_to_line_indices_dict[token]
            if len(lines_to_consider) > self._max_lines_per_token:
                lines_to_consider = random.sample(lines_to_consider, self._max_lines_per_token)

            for line_index in lines_to_consider:
                params.append((self._model, line_index, \
                    self._corpus_lines, self._model_encoded_corpus_lines_token_ids, self._model_encoded_corpus_lines_token_pieces, \
                    token, current_vocab, \
                    self._target_embeddings, self._context_embeddings, self._log, self._window_size))

            lines_per_token[token] = len(lines_to_consider)
        
        res = process_pool.starmap(Utils.get_diff_sg_wo_token_for_line, params)

        for r in res:
            token = r[0]
            sg_wo_diff = r[1]

            if token not in token_and_sg_log_prob_without_it.keys():
                token_and_sg_log_prob_without_it[token] = current_total_sg

            token_and_sg_log_prob_without_it[token] += (float(sg_wo_diff) / lines_per_token[token])
            
        return token_and_sg_log_prob_without_it

    def get_model(self):
        return self._model

    def set_vocab(self, new_vocabulary):
        # Note: this indeed changing the model vocab, but not model.vocab_size().
        # if you want to test this - remove some token from the vocabulary, and compare encodings before and after removing.
        self._model.set_vocabulary(new_vocabulary)
        self._current_vocab = new_vocabulary
        self._should_compute_vocab = False
        



# Copyright © 2023 Kensho Technologies, LLC

class Word2VecParams:
    def __init__(self, D, N, ALPHA, window_size, min_count, sg):
        self.D = D
        self.N = N
        self.ALPHA = ALPHA
        self.window_size = window_size
        self.min_count = min_count
        self.sg = sg



# -*- coding: utf-8 -*-

import multiprocessing as mp
import sys
import json
import random
import pickle
import argparse
import numpy as np
from tqdm import tqdm
import sentencepiece as spm
from os.path import exists
import os

# Import From Python-Modules
sys.path.insert(0, "Python-Modules")
import Utils
import Corpus
import Logger
import SG_BPE
import Embeddings

######### Fix Random Seed ###########################################################

# Fill with your chosen seed
CHOSEN_SEED = 0

def set_random_seed(chosen_seed):
	# setting seed
	random.seed(chosen_seed)
	spm.SetRandomGeneratorSeed(chosen_seed)
	np.random.seed(chosen_seed)

def fix_random_seed(experiment_name, is_continue_execution, chosen_seed, log):
	seed_filepath = "results/{}/seed.txt".format(experiment_name)
	if is_continue_execution and exists(seed_filepath):
		with open(seed_filepath, "r") as seed_file:
			chosen_seed = int(seed_file.read())
			log.info("chosen seed is {}".format(chosen_seed))
	
	else:
		# saving for continue execution of that experiment, if this is first time execution
		with open(seed_filepath, "w+") as seed_file:
			seed_file.write(str(chosen_seed))

	set_random_seed(chosen_seed)

######### Log Parameters #####################################################################

def log_parameters(log, final_vocab_size, initial_vocab_size, \
	partial_corpus_lines_number, tokens_to_prune_in_iteration, tokens_to_consider_in_iteration, \
	iterations_until_reranking, max_lines_per_token, corpus_filepath, partial_corpus_filepath, window_size):
	
	log.info("-----------------------------------------")
	log.info("Starting Experiment.")
	log.info("initial_vocab_size: {}, final_vocab_size: {}".format(initial_vocab_size, final_vocab_size))
	log.info("partial_corpus_lines_number: {}\n".format(partial_corpus_lines_number))
	log.info("tokens_to_prune_in_iteration {}".format(tokens_to_prune_in_iteration))
	log.info("tokens_to_consider_in_iteration: {}".format(tokens_to_consider_in_iteration))
	log.info("iterations_until_reranking: {}".format(iterations_until_reranking))
	log.info("max_lines_per_token: {}".format(max_lines_per_token))
	log.info("corpus_filepath: {}".format(corpus_filepath))
	log.info("partial_corpus_filepath: {}".format(partial_corpus_filepath))
	log.info("window_size: {}".format(window_size))
	log.info("Number of processes in pool: {}".format(mp.cpu_count()))
	log.info("Not re-calculating embeddings")
	log.info("-----------------------------------------")

######### Main #####################################################################

def main(experiment_name, is_continue_execution, final_vocab_size, \
	initial_vocab_size, partial_corpus_lines_number, tokens_to_prune_in_iteration, tokens_to_consider_in_iteration, \
	iterations_until_reranking, max_lines_per_token, corpus_filepath, partial_corpus_filepath, window_size):

	# Initialize Statistics Logger
	print("Initializing statistics logger")
	log = Logger.Logger("statistics")

	if not exists("results"):
		log.info("Creating 'results' dir")
		os.mkdir("results")

	experiment_results_directory = "results/{}".format(experiment_name)
	if not exists(experiment_results_directory):
		log.info("Creating {} dir".format(experiment_results_directory))
		os.mkdir(experiment_results_directory)

	log.info("Fixing random seed")
	fix_random_seed(experiment_name, is_continue_execution, CHOSEN_SEED, log)

	# Preparing Data
	log.info("Preparing corpus")
	corpus = Corpus.Corpus(corpus_filepath, partial_corpus_filepath, partial_corpus_lines_number, log)
	partial_corpus = corpus.get_corpus()

	# Log parameters if not continuation execution
	if not is_continue_execution:
		log_parameters(log, final_vocab_size, initial_vocab_size, partial_corpus_lines_number, \
			tokens_to_prune_in_iteration, tokens_to_consider_in_iteration, iterations_until_reranking, \
			max_lines_per_token, corpus_filepath, partial_corpus_filepath, window_size)

	# BPE using SentencePiece Python Module
	log.info("Preparing SG and BPE Models")
	models = SG_BPE.SG_BPE_Models(experiment_name, is_continue_execution, final_vocab_size, initial_vocab_size, partial_corpus_filepath)
	sg_bpe_model = models.get_sg_bpe_model()
	bpe_vanilla_model = models.get_bpe_vanilla_model()

	# Computing Embedding Matrix
	log.info("Computing embeddings")
	embeddings_filepath = "results/{}/embeddings.bin".format(experiment_name)
	if not is_continue_execution:
		wp_trainer = Embeddings.EmbeddingsTrainer(sg_bpe_model, corpus, window_size, log)
		target_embeddings, context_embeddings = wp_trainer.train_embeddings()
		with open(embeddings_filepath, "wb") as embeddings_file:
			pickle.dump(target_embeddings, embeddings_file)
			pickle.dump(context_embeddings, embeddings_file)
			log.info("dumped embeddings to {}".format(embeddings_filepath))
	else:
		with open(embeddings_filepath, "rb") as embeddings_file:
			target_embeddings = pickle.load(embeddings_file)
			context_embeddings = pickle.load(embeddings_file)

	# Creating model objects
	# These will hold needed logic for training loop, logging results and gathering information about models too
	log.info("Creating model objects")
	vocab_filepath = "results/{}/current_vocab".format(experiment_name)
	sg_bpe_model_object = SG_BPE.Model(experiment_name, log, sg_bpe_model, "sg_bpe", \
		target_embeddings, context_embeddings, partial_corpus, max_lines_per_token, window_size, is_continue_execution, vocab_filepath)
	bpe_vanilla_model_object = SG_BPE.Model(experiment_name, log, bpe_vanilla_model, "bpe_vanilla", \
		target_embeddings, context_embeddings, partial_corpus, max_lines_per_token, window_size)

	# We use the "get_current_vocab" method because of sentencepiece tricky way to remove tokens from the vocabulary.
	current_vocab = sg_bpe_model_object.get_current_vocab()
	with open(vocab_filepath + ".bin", "wb") as vocab_file:
		pickle.dump(current_vocab, vocab_file)

	###########################################################################################################################
	# Prune Tokens - SkipGram Log Probability
	# For now we compute SG objective without negative samples - this causes noise, and due to the fact we look at the actual result 
	# 	(and don't just refer it as objective) we don't want that noise.
	# We say we have better vocabulary when we minimize this objective.
	# Thus, in each iteration we will find the tokens that without them the objective is minimal.
	###########################################################################################################################

	log.info("Preparing sorted Tokens-SG-objective list")
	sorted_tokens_sg_filepath = "results/{}/sorted_tokens_sg.bin".format(experiment_name)
	if is_continue_execution and exists(sorted_tokens_sg_filepath):
		with open(sorted_tokens_sg_filepath, "rb") as sorted_tokens_sg_file:
			sorted_tokens_sg = pickle.load(sorted_tokens_sg_file)
	else:
		# Log starting-point vocabulary:	
		sg_bpe_model_object.log_experiments_model_results(partial_corpus_filepath)
		bpe_vanilla_model_object.log_experiments_model_results(partial_corpus_filepath)

		# And Log starting-point SG objective
		total_skipgram_ns_probability = sg_bpe_model_object.total_sg_log_prob(partial_corpus_filepath)
		log.info("Initial SG-BPE total_log_sg_prob: {}".format(total_skipgram_ns_probability))

		bpe_vanilla_total_skipgram_ns_probability = bpe_vanilla_model_object.total_sg_log_prob(partial_corpus_filepath)
		log.info("Initial Vanilla-BPE total_log_sg_prob: {}".format(bpe_vanilla_total_skipgram_ns_probability))

		log.log_separator()

		# Log original sg log prob without each token
		original_token_and_sg_log_prob_without_it = sg_bpe_model_object.get_sg_log_prob_without_tokens_mp(total_skipgram_ns_probability, partial_corpus_filepath, True)
		sorted_tokens_sg = sorted(original_token_and_sg_log_prob_without_it.items(), key=lambda item: item[1])
		sg_log_prob_without_tokens_mp_filepath = "./results/{}/sg_log_probs_without_tokens_mp.txt".format(experiment_name)
		with open(sg_log_prob_without_tokens_mp_filepath, "w+") as logprob_file:
			logprob_file.write(json.dumps(sorted_tokens_sg, indent=4))

		with open(sorted_tokens_sg_filepath, "wb") as sorted_tokens_sg_file:
			pickle.dump(sorted_tokens_sg, sorted_tokens_sg_file)

	# And start the "Training-Loop"
	log.info("Starting the 'Training-Loop'")

	current_vocab = sg_bpe_model_object.get_current_vocab()
	current_total_sg = sg_bpe_model_object.total_sg_log_prob(partial_corpus_filepath)

	## prepare from ahead ####
	### now we want dict from token to index of lines - and re-encode when token is ablated
	sg_bpe_model_object.initialize_encoded_form_for_corpus_lines()
	sg_bpe_model_object.initialize_token_to_line_indices_dictionary(current_vocab, partial_corpus, experiment_name, is_continue_execution)

	# loop and ablate
	iteration = 0
	current_dict_of_top_tokens = {}

	progress_bar = tqdm(total = (len(current_vocab) - final_vocab_size) / tokens_to_prune_in_iteration)
	while len(current_vocab) > final_vocab_size:
		log.info("iteration #{}, num of vocab: {}, length of top: {}".format(iteration, len(current_vocab), len(current_dict_of_top_tokens)))

		if (iteration % iterations_until_reranking) == 0:
			token_and_sg_log_prob_without_it = sg_bpe_model_object.get_sg_log_prob_without_tokens_mp2(current_total_sg, True)
			sorted_tokens_sg = sorted(token_and_sg_log_prob_without_it.items(), key=lambda item: item[1])
			current_dict_of_top_tokens = [t[0] for t in sorted_tokens_sg[:tokens_to_consider_in_iteration]]
		else:
			token_and_sg_log_prob_without_it = sg_bpe_model_object.get_sg_log_prob_without_tokens_mp2(current_total_sg, True, current_dict_of_top_tokens)
			sorted_tokens_sg = sorted(token_and_sg_log_prob_without_it.items(), key=lambda item: item[1])

		# finding the tokens to prune in that iteration (those that without them the value is minimal)
		tokens_to_prune = [t[0] for t in sorted_tokens_sg[:tokens_to_prune_in_iteration]]
		log.info("pruning: {}".format(tokens_to_prune))
		
		for t in tokens_to_prune:
			try:
				current_vocab.remove(t)
				current_dict_of_top_tokens.remove(t)
			except:
				log.info("could not remove {}".format(t[0]))
				raise

		# update our model vocab
		with open(vocab_filepath + ".bin", "wb") as vocab_file:
			pickle.dump(current_vocab, vocab_file)

		sg_bpe_model_object.set_vocab(current_vocab)
		sg_bpe_model_object.update_encoded_form_for_corpus_lines(tokens_to_prune)
		
		# upgrade our current total sg
		current_total_sg = sg_bpe_model_object.total_sg_log_prob(partial_corpus_filepath)

		# log current sg objective
		log.info("current SG-BPE total_log_sg_prob: {}".format(current_total_sg))
		log.info("current size is {}".format(len(current_vocab)))
		log.info("Time elapsed (iteration #{}) - {} (minutes)".format(iteration, float(progress_bar.format_dict["elapsed"])/60))
		log.info("Time elapsed (iteration #{}) - {} (minutes)".format(iteration, float(progress_bar.format_dict["elapsed"])/60))

		# prepare to next iteration
		log.log_separator()
		progress_bar.update(1)
		iteration += 1

	progress_bar.close()

	# Logging experiment results:
	sg_bpe_model_object.log_experiments_model_results(partial_corpus_filepath, "sg_bpe")

	final_total_skipgram_ns_probability = sg_bpe_model_object.total_sg_log_prob(partial_corpus_filepath)
	log.info("Final SG-BPE total_log_sg_prob: {}".format(final_total_skipgram_ns_probability))

	final_bpe_vanilla_total_skipgram_ns_probability = bpe_vanilla_model_object.total_sg_log_prob(partial_corpus_filepath)
	log.info("Vanilla-BPE total_log_sg_prob: {}".format(final_bpe_vanilla_total_skipgram_ns_probability))

	# Log difference between vanilla and updated bpe models
	# log tokens only in sg-bpe vocab
	current_vocab = sg_bpe_model_object.get_current_vocab()
	log.info("vocab len = {}".format(len(current_vocab)))
	bpe_vanilla_vocab = bpe_vanilla_model_object.get_current_vocab()
	log.info("bpe vanilla vocab len = {}".format(len(bpe_vanilla_vocab)))

	only_in_sg_bpe = [t for t in current_vocab if t not in bpe_vanilla_vocab]
	sg_bpe_only_filepath = "./results/{}/sg_bpe_only.txt".format(experiment_name)
	with open(sg_bpe_only_filepath, "w+") as sg_bpe_only:
		sg_bpe_only.write(json.dumps(only_in_sg_bpe, indent=4))

	# log tokens only in bpe-vanilla vocab
	only_in_bpe_vanilla = [t for t in bpe_vanilla_vocab if t not in current_vocab]
	bpe_vanilla_only_filepath = "./results/{}/bpe_vanilla_only.txt".format(experiment_name)
	with open(bpe_vanilla_only_filepath, "w+") as bpe_vanilla_only:
		bpe_vanilla_only.write(json.dumps(only_in_bpe_vanilla, indent=4))
		
	# Average token length
	average_token_length = sum(map(len, current_vocab)) / len(current_vocab)
	log.info("SG-BPE average token length: {}".format(average_token_length))

	# Average token length of vanilla-bpe
	vanilla_average_token_length = sum(map(len, bpe_vanilla_vocab)) / len(bpe_vanilla_vocab)
	log.info("Vanilla-BPE average token length: {}".format(vanilla_average_token_length))

	# Length of encoded file data
	with open(partial_corpus_filepath) as input_file:
		input_file_data = input_file.read()

	encoded_data = [sg_bpe_model.id_to_piece(x) for x in sg_bpe_model.encode(input_file_data)]
	log.info("SG-BPE encoding length: {} (file length: {})".format(len(encoded_data), len(input_file_data)))

	# Length of encoded file data of vanilla-bpe
	vanilla_encoded_data = [bpe_vanilla_model.id_to_piece(x) for x in bpe_vanilla_model.encode(input_file_data)]
	log.info("Vanilla-BPE encoding length: {} (file length: {})".format(len(vanilla_encoded_data), len(input_file_data)))

def prepare_parameters():
	parser = argparse.ArgumentParser(description="Calculating SaGe vocabulary.")
	parser.add_argument("experiment_name", help="name of experiment, will save results under that name.") # Required param (no "--" prefix)
	parser.add_argument("--final_vocab_size", required=True, help="final vocabulary size")
	parser.add_argument("--initial_vocab_size", required=True, help="initial vocabulary size")
	parser.add_argument("--tokens_to_prune_in_iteration", required=True, help="number of tokens to prune in each iteration")
	parser.add_argument("--tokens_to_consider_in_iteration", required=True, help="number of tokens to consider in each iteration")
	parser.add_argument("--iterations_until_reranking", required=True, help="number of iterations until reranking")
	parser.add_argument("--is_continue", default="N", help="is this execution continues former execiution of that experiment: [Y/N]")
	parser.add_argument("--corpus_filepath", required=True, help="filepath for full corpus (like wiki corpus)")
	parser.add_argument("--thousands_of_corpus_lines", default=200, help="number of corpus lines - in thousands")
	parser.add_argument("--partial_corpus_filepath", required=True, help="where to create partial corpus file - with number of lines requested")
	parser.add_argument("--max_lines_per_token", default=1000, help="max number of lines to consider in objective calculation, per-token")
	parser.add_argument("--window_size", default=5, help="window size for SG objective calculation, and also for embeddings calculation")
	return vars(parser.parse_args())

if __name__ == "__main__":
	args = prepare_parameters()

	experiment_name = args["experiment_name"]
	print("Starting experiment {}".format(experiment_name))

	is_continue_execution = True if args["is_continue"] == "Y" else False
	print("is_continue={}".format(is_continue_execution))
	
	partial_corpus_lines_number = int(args["thousands_of_corpus_lines"]) * 1000
	final_vocab_size = int(args["final_vocab_size"])
	initial_vocab_size = int(args["initial_vocab_size"])
	tokens_to_prune_in_iteration = int(args["tokens_to_prune_in_iteration"])
	tokens_to_consider_in_iteration = int(args["tokens_to_consider_in_iteration"])
	iterations_until_reranking = int(args["iterations_until_reranking"])
	max_lines_per_token = int(args["max_lines_per_token"])
	corpus_filepath = args["corpus_filepath"]
	partial_corpus_filepath = args["partial_corpus_filepath"]
	window_size = int(args["window_size"])

	main(experiment_name, is_continue_execution, final_vocab_size, \
		initial_vocab_size, partial_corpus_lines_number, tokens_to_prune_in_iteration, tokens_to_consider_in_iteration, \
		iterations_until_reranking, max_lines_per_token, corpus_filepath, partial_corpus_filepath, window_size)



# Copyright © 2023 Kensho Technologies, LLC

import logging

import sage_tokenizer.embeddings as emb
import sage_tokenizer.utils as utils
from sage_tokenizer.model import SaGeTokenizer
from sage_tokenizer.Word2VecParams import Word2VecParams


class SaGeVocabBuilder:

    def __init__(self, full_vocab_schedule: list[int], embeddings_schedule: list[int], max_len=16, workers_number=1,
                 random_seed=1234, word2vec_d=50, word2vec_n=15, word2vec_alpha=0.025, word2vec_window_size=5,
                 word2vec_min_count=1, word2vec_sg=1):
        self.full_vocab_schedule = full_vocab_schedule
        self.embeddings_schedule = embeddings_schedule
        self.max_len = max_len
        self.workers_number = workers_number
        self.random_seed = random_seed
        self.word2vec_params = Word2VecParams(
            D=word2vec_d,
            N=word2vec_n,
            ALPHA=word2vec_alpha,
            window_size=word2vec_window_size,
            min_count=word2vec_min_count,
            sg=word2vec_sg
        )

    def build_vocab(self, experiment_name: str, corpus_filepath: str, vocabulary_filepath: str,
                    partial_corpus_filepath="", partial_corpus_line_number=1000):

        utils.init_logger(experiment_name)
        logging.info(f"Start experiment {experiment_name}")
        logging.info(f"Process will use up to {self.workers_number} worker threads.")

        logging.info("Getting output directories")
        embeddings_folder, stats_folder, vocab_folder = utils.get_output_folder(experiment_name)
        logging.info("Setting random seed")
        utils.set_random_seed(experiment_name, self.random_seed)
        logging.info(f"Loading initial vocabulary from file {vocabulary_filepath}")
        byte_vocab = utils.load_vocab(vocabulary_filepath)
        logging.info(f"Finished loading initial vocabulary. Vocabulary size: {len(byte_vocab)}")

        actual_max_len = max([len(v) for v in byte_vocab])
        if self.max_len != actual_max_len:
            logging.warning(f"max_len parameter value {self.max_len} doesn't match actual max {actual_max_len}")

        logging.info("Initializing tokenizer")
        sage_model = SaGeTokenizer(byte_vocab, self.max_len)

        logging.info(f"Loading Corpus from {corpus_filepath}")
        partial_corpus = utils.load_corpus(corpus_filepath, partial_corpus_filepath, partial_corpus_line_number)
        logging.info("Starting the training loop")
        vocab_schedule = self.full_vocab_schedule

        if not len(vocab_schedule) >= 2:
            raise Exception("Vocabulary schedule must contain more than 2 vocabulary sizes!")

        vocab_schedule.sort(reverse=True)  # largest first
        logging.info(f"initial vocab_schedule is {vocab_schedule[0]} vs actual size {sage_model.vocab_size()}")

        embedding_sizes = set(self.embeddings_schedule)

        # initialize embeddings for first iteration
        embeddings = emb.get_embeddings(vocab_schedule[0], embeddings_folder, partial_corpus, sage_model,
                                        self.workers_number, self.word2vec_params)

        # skipping the initial vocab size here
        i = 0
        # stop one before the end, since we do i+1,
        # so we'll make the vocab of that final size, but won't do the tokenization on it
        while i < len(vocab_schedule) - 1:

            current_step_vocab_size = vocab_schedule[i]  # this will be the label used for files
            target_vocab_size = vocab_schedule[i + 1]
            actual_vocab_size = sage_model.vocab_size()
            logging.info(f"\nRound {i} - Start: "
                         f"\n\tCurrent step vocabulary size: {current_step_vocab_size}, "
                         f"\n\tTarget vocabulary size: {target_vocab_size}, "
                         f"\n\tActual vocabulary size: {actual_vocab_size}")

            if vocab_schedule[i] in embedding_sizes:
                embeddings = emb.get_embeddings(current_step_vocab_size, embeddings_folder, partial_corpus, sage_model,
                                                self.workers_number, self.word2vec_params)

            if actual_vocab_size <= target_vocab_size:
                logging.info(f"Actual vocab is already smaller than target. continue to next iteration ")
                i += 1
                continue

            # call sage in parallel
            logging.info(f"Sage started.")
            total_tokens, total_triples, token_to_losses, ablated_sizes = utils.run_sage_parallel(embeddings,
                                                                                                  partial_corpus,
                                                                                                  sage_model,
                                                                                                  self.workers_number)
            logging.info(f"Sage finished. total tokens: {total_tokens}, total triplets: {total_triples}")

            # token_to_losses won't include any single byte tokens, but we want to keep those around
            # so lets just add them with large scores, so they stay around
            vocab_size_before_single_byte_tokens_addition = len(token_to_losses)
            sage_model.add_all_byte_ids(token_to_losses, score=1e6)
            logging.info(f"Adding single bytes to vocab. Size before: {vocab_size_before_single_byte_tokens_addition}, "
                         f"size after: {len(token_to_losses)}")

            # if a token doesn't appear in token_to_losses then it didn't participate in the tokenization
            current_active_vocab_size = len(token_to_losses)
            current_inactive_vocab_size = actual_vocab_size - len(token_to_losses)
            logging.info(f"Actual vocab size: {actual_vocab_size}, "
                         f"Target vocab size: {target_vocab_size}, "
                         f"Active Vocab Size: {current_active_vocab_size}, "
                         f"Inactive Vocab Size: {current_inactive_vocab_size}")

            # how many of the losses are negative
            neg_loss = len([loss for loss in token_to_losses.values() if loss < 0.0])
            zero_loss = len([loss for loss in token_to_losses.values() if loss == 0.0])
            pos_loss = len([loss for loss in token_to_losses.values() if loss > 0.0])
            logging.info(f"Negative losses: {neg_loss}, zero losses: {zero_loss}, positive losses: {pos_loss}")

            # in case the active vocab we found is actually smaller than the target vocab,
            # change the target to the next one, until it's smaller than the vocab we found,
            # so the ablation part will actually do something
            while current_active_vocab_size <= target_vocab_size:
                logging.info(f"Active vocab size is {current_active_vocab_size} - "
                             f"smaller than target {target_vocab_size}. Moving to next target_vocab_size"
                             f"\n\n(Round number increased to {i + 1})\n")
                i += 1
                target_vocab_size = vocab_schedule[i + 1]
                logging.info(f"New target_vocab_size: {target_vocab_size}")

            num_tokens_to_prune = current_active_vocab_size - target_vocab_size
            logging.info(f"Num tokens to prune {num_tokens_to_prune}")

            ######################
            # do the ablation
            ######################
            # we want to drop the smallest (negative) values
            # these are the ones with the largest decrease in likelihood from dropping the ablated token
            sorted_losses = list(sorted([(loss, tid) for (tid, loss) in token_to_losses.items()]))
            utils.save_sorted_losses(sage_model, sorted_losses, target_vocab_size, vocab_folder)

            stats = {
                "current_step_vocab_size": current_step_vocab_size, "total_tokens": total_tokens,
                "total_triples": total_triples, "current_active_vocab_size": current_active_vocab_size,
                "current_inactive_vocab_size": current_inactive_vocab_size, "neg_loss": neg_loss,
                "zero_loss": zero_loss, "pos_loss": pos_loss, "target_vocab_size": target_vocab_size,
                "num_tokens_to_prune": num_tokens_to_prune, "ablated_sizes": ablated_sizes,
            }
            utils.save_stats(stats, stats_folder, target_vocab_size)

            # these are the tokens to be removed
            tokens_to_prune = set([sage_model.id_to_bytes(tid) for (loss, tid) in sorted_losses[:num_tokens_to_prune]])
            # double check there are no single bytes tokens to prune here
            single_byte_tokens_to_prune = [token for token in tokens_to_prune if len(token) == 1]
            assert len(single_byte_tokens_to_prune) == 0

            # our active vocabulary *after* pruning
            # is active if it has an entry in token_to_losses
            active_vocab = {tok: tid for (tok, tid) in sage_model.get_vocabulary().items()
                            if tid in token_to_losses and tok not in tokens_to_prune}

            # our overall vocabulary after pruning
            target_vocab = {tok: tid for (tok, tid) in sage_model.get_vocabulary().items()
                            if tok not in tokens_to_prune}

            # the deleted items
            deleted_vocab = {tok: tid for (tok, tid) in sage_model.get_vocabulary().items()
                             if tok in tokens_to_prune}

            vocab_save_name = f"{vocab_folder}/sage_vocab_{target_vocab_size}.vocab"
            logging.info(f"Saving intermediate vocab of size {len(target_vocab)} to {vocab_save_name}")
            utils.write_vocab(target_vocab, vocab_save_name)

            active_save_name = f"{vocab_folder}/active_vocab_{target_vocab_size}.vocab"
            logging.info(f"Saving active vocab of size {len(active_vocab)} to {active_save_name}")
            utils.write_vocab(active_vocab, active_save_name)

            # save the deleted ones too for analysis, with the original size
            deleted_save_name = f"{vocab_folder}/deleted_vocab_{target_vocab_size}.vocab"
            logging.info(f"Saving deleted vocab of size {len(deleted_vocab)} to {deleted_save_name}")
            utils.write_vocab(deleted_vocab, deleted_save_name)

            # now update the internal state of sage_model to use the new smaller vocab
            # pass in list of bytes keys, which keep insertion order
            sage_model.set_vocabulary(list(target_vocab.keys()))

            logging.info(f"\nRound {i} - End: "
                         f"\n\tCurrent step vocabulary size: {current_step_vocab_size}, "
                         f"\n\tTarget vocabulary size: {target_vocab_size}, "
                         f"\n\tActual vocabulary size:{len(active_vocab)}")

            # advance to next smaller size
            i += 1



import logging

logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
    level=logging.INFO,
)

class Logger:
    def __init__(self, name):
        self.logger = logging.getLogger(name)

    def info(self, message, *args, **kwargs):
        self.logger.info(message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        self.logger.warn(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        self.logger.error(message, *args, **kwargs)

    def log_separator(self):
        self.logger.info("---------------------")




from scipy.special import expit
import numpy as np
import math

SPECIAL_TOKENS = ["<unk>", "<s>", "</s>"]
WORD_PREFIX_CHAR = "\u2581"

def sigmoid(num):
    s = expit(num)
    if s == 1.0:
        raise BaseException("sigmoid overflow")

    return expit(num)

def compute_window(token_index, tokens_in_line, window_size):
    # get context window tokens
    context_start = max(token_index-window_size, 0)
    context_end = min(token_index+window_size+1, len(tokens_in_line))
    window = tokens_in_line[context_start:token_index] + tokens_in_line[token_index+1:context_end]

    return window, context_start, context_end

# Compose list of tokens we never ablate
def get_not_ablateable_tokens_list(current_vocab):
    NAT_list = []
    for t in current_vocab:
        if len(t) == 1:
            NAT_list.append(t)
            continue

        if t.startswith(WORD_PREFIX_CHAR) and len(t) == 2:
            NAT_list.append(t)
            continue

        if t in SPECIAL_TOKENS:
            NAT_list.append(t)
            continue

        continue
        
    return NAT_list

# calculating 'offset' = number of tokens added because of our new encoding
def calculate_token_offset(i, last_times_accumulate_offset, tokens_in_line, updated_tokens_in_line, log):
    # original encoding next_word_index
    original_current_index = i+1
    if original_current_index == len(tokens_in_line):
        original_current_index = i
    else:
        while not tokens_in_line[original_current_index].startswith("\u2581"):
            original_current_index += 1
            if original_current_index == len(tokens_in_line):
                break
    
    original_next_word_index = original_current_index

    # and the updated next_word_index
    current_index = last_times_accumulate_offset + i+1
    if current_index == len(updated_tokens_in_line):
        current_index = last_times_accumulate_offset + i
    else:
        if (current_index >= len(updated_tokens_in_line)):
            log.info("Fail! current index = {}, length of \"updated_tokens_in_line\" = {}".format(current_index, len(updated_tokens_in_line)))
            log.info("token #{}, updated tokens:{}".format(i, updated_tokens_in_line))
            current_index -= 1
        else:
            while not updated_tokens_in_line[current_index].startswith("\u2581"):
                current_index += 1
                if current_index == len(updated_tokens_in_line):
                    break

    updated_next_word_index = current_index

    # the offset is the difference between them
    offset = updated_next_word_index - original_next_word_index - last_times_accumulate_offset

    return offset

def token_to_line_indices_dictionary(current_vocab, corpus_lines):
    # first lets hold the instance of line with "/u2581" chars instead of spaces
    corpus_lines_encoded = []
    for line in corpus_lines:
        enc_line = WORD_PREFIX_CHAR + line
        enc_line = enc_line.replace(" ", WORD_PREFIX_CHAR)
        corpus_lines_encoded.append(enc_line)

    # and test whether tokens contained in the lines
    token_to_line_dict = {}
    for token in current_vocab:
        lines_with_token_indices = []
        for i, enc_line in enumerate(corpus_lines_encoded):
            if token not in enc_line:
                continue
            lines_with_token_indices.append(i)
        token_to_line_dict[token] = lines_with_token_indices

    return token_to_line_dict

####################################################################
## Functions for multiprocessing - cannot be class methods #########
####################################################################

def sg_for_window_mp(target_token, window, target_embeddings, context_embeddings, log):
    current_p = 0
    for w in window:
        # calculate current value to add
        dot_product = np.dot(target_embeddings[target_token], context_embeddings[w])
        try:
            current_p += math.log(sigmoid(dot_product))
        except:
            pass

    return (-1) * current_p


def substract_windows_from_sg_mp(i, tokens_in_line_ints, current_sg, target_embeddings, context_embeddings, window_size, log):
    # we should substract windows of all tokens in our original window
    sg_wo = current_sg
    _, context_start, context_end = compute_window(i, tokens_in_line_ints, window_size)

    for index in range(context_start, context_end):
        window, _, _ = compute_window(index, tokens_in_line_ints, window_size)
        sg_wo -= sg_for_window_mp(tokens_in_line_ints[index], window, target_embeddings, context_embeddings, log)

    return sg_wo

def add_windows_to_sg_mp(model, updated_i, offset, updated_context_start, \
                        updated_context_end, updated_tokens_in_line, current_sg, \
                        target_embeddings, context_embeddings, window_size, log):
    sg_wo = current_sg
    
    # assume it now looks like: old[context_start:i] + [ablated_token_encoding] + old[i+1:context_end]
    # add the windows before i (ablated token)
    for index in range(updated_context_start, updated_i):
        try:
            window, _, _ = compute_window(index, updated_tokens_in_line, window_size)
            sg_wo += sg_for_window_mp(updated_tokens_in_line[index], window, target_embeddings, context_embeddings, log)
        except:
            print("index  {}".format(index))
            print("tokens {}".format(updated_tokens_in_line))
            raise

    # add the ablated token encoding windows
    for index in range(updated_i, updated_i+offset):
        try:
            window, _, _ = compute_window(index, updated_tokens_in_line, window_size)
            sg_wo += sg_for_window_mp(updated_tokens_in_line[index], window, target_embeddings, context_embeddings, log)
        except:
            print("index  {}".format(index))
            print("tokens {}".format(updated_tokens_in_line))
            raise

    # add the windows after the ablated token
    for index in range(updated_i+offset, updated_context_end):
        try:
            window, _, _ = compute_window(index, updated_tokens_in_line, window_size)
            sg_wo += sg_for_window_mp(updated_tokens_in_line[index], window, target_embeddings, context_embeddings, log)
        except:
            print("index  {}, len {}".format(index, len(updated_tokens_in_line)))
            updated_tokens_in_line_str = [model.id_to_piece(x) for x in updated_tokens_in_line]
            print("tokens {}".format(updated_tokens_in_line_str))
            raise

    return sg_wo

def update_sg_per_instance_of_token_mp(model, token_to_ablate, i, \
                                    line, tokens_in_line_ints, tokens_in_line_pieces, \
                                    last_times_accumulate_offset, \
                                    current_total_sg, current_vocab, \
                                    target_embeddings, context_embeddings, log, window_size):
    
    # window of token - before ablation
    sg_wo = current_total_sg

    # we should substract windows of all tokens in our original window
    original_window, original_context_start, original_context_end = compute_window(i, tokens_in_line_pieces, window_size)

    sg_wo = substract_windows_from_sg_mp(i, tokens_in_line_ints, sg_wo, target_embeddings, context_embeddings, window_size, log)

    ## remove token from current vocab
    vocab_without_token = current_vocab.copy()
    vocab_without_token.remove(token_to_ablate)
    model.set_vocabulary(vocab_without_token)

    # encode with new vocab
    updated_tokens_in_line_ints = model.encode(line, out_type=int)
    updated_tokens_in_line = [model.id_to_piece(u) for u in updated_tokens_in_line_ints]

    # calculating 'offset' = number of tokens added because of our new encoding
    offset = calculate_token_offset(i, last_times_accumulate_offset, tokens_in_line_pieces, updated_tokens_in_line, log)

    #################################################################################
    # make sure the window (before the ablated token) stays the same ################
    #################################################################################
    _, updated_context_start, updated_context_end = compute_window(i+last_times_accumulate_offset, updated_tokens_in_line_ints, window_size)
    updated_context_end = min(updated_context_end+offset, len(updated_tokens_in_line)) # real context_en
    updated_i = i + last_times_accumulate_offset
    updated_window = updated_tokens_in_line[updated_context_start:updated_i] + updated_tokens_in_line[updated_i+offset+1:updated_context_end]

    first_part_of_window_length = min(window_size, updated_i)

    if updated_window[:first_part_of_window_length] != original_window[:first_part_of_window_length]:
        return current_total_sg, offset

    #################################################################################
    #################################################################################

    # and add the updated windows sg
    sg_wo = add_windows_to_sg_mp(model, updated_i, offset, updated_context_start, \
                                updated_context_end, updated_tokens_in_line_ints, \
                                sg_wo, target_embeddings, context_embeddings, window_size, log)

    ## revert model to current vocab
    model.set_vocabulary(current_vocab)

    return sg_wo, offset

## We assume the line contains the token_to_ablate (check it before shooting new process to execute this method)
def get_diff_sg_wo_token_for_line(model, line_index, \
                                    corpus_lines, model_encoded_corpus_lines_token_ids, model_encoded_corpus_lines_token_pieces, \
                                    token_to_ablate, current_vocab, target_embeddings, context_embeddings, log, window_size):
    sg_wo_diff = 0

    tokens_in_line_ints = model_encoded_corpus_lines_token_ids[line_index]
    tokens_in_line_pieces = model_encoded_corpus_lines_token_pieces[line_index]
    line = corpus_lines[line_index]
    
    # For start, assume there are no overlaps in the windows of 2 instances
    indices = [i for i, x in enumerate(tokens_in_line_pieces) if x == token_to_ablate]

    last_time_offset = 0
    last_times_accumulate_offset = 0
    for i in indices:
        sg_wo_diff, last_time_offset = update_sg_per_instance_of_token_mp(model, token_to_ablate, i, line, tokens_in_line_ints, tokens_in_line_pieces, last_times_accumulate_offset, sg_wo_diff, current_vocab, target_embeddings, context_embeddings, log, window_size)
        last_times_accumulate_offset += last_time_offset

    return token_to_ablate, sg_wo_diff

def sg_wo_token_mp(model, token_to_ablate, current_total_sg, current_vocab, training_filepath, target_embeddings, context_embeddings, log, corpus_lines, window_size):
    sg_wo = current_total_sg

    for line in corpus_lines:
        tokens_in_line_pieces = [model.id_to_piece(x) for x in model.encode(line, out_type=int)]
        if token_to_ablate not in tokens_in_line_pieces:
            continue

        # For start, assume there are no overlaps in the windows of 2 instances
        indices = [i for i, x in enumerate(tokens_in_line_pieces) if x == token_to_ablate]

        last_time_offset = 0
        last_times_accumulate_offset = 0
        for i in indices:
            tokens_in_line_ints = [model.piece_to_id(t) for t in tokens_in_line_pieces]
            sg_wo, last_time_offset = update_sg_per_instance_of_token_mp(model, token_to_ablate, i, line, tokens_in_line_ints, tokens_in_line_pieces, last_times_accumulate_offset, sg_wo, current_vocab, target_embeddings, context_embeddings, log, window_size)
            last_times_accumulate_offset += last_time_offset

    return token_to_ablate, sg_wo



# Copyright © 2023 Kensho Technologies, LLC

import numpy as np

from sage_tokenizer.HFEncoding import HFEncoding
from sage_tokenizer.utils import verify_all_single_byte_exist_in_vocab


class SaGeTokenizer:

    def __init__(self, initial_vocabulary, max_len=16):
        self.hfe = HFEncoding()
        self.byte_vocab = None
        self.inv_byte_vocab = None
        self.str_vocab = None
        self.inv_str_vocab = None

        self.set_vocabulary(initial_vocabulary)
        self.max_len = max_len

    # given an order list of bytes for vocabulary
    # create our internal structures
    # overwriting any previous values
    def set_vocabulary(self, new_vocab: list[bytes]):
        self.byte_vocab = self.set_bytes_vocab(new_vocab)

        # make sure we always have all single bytes in vocabulary
        verify_all_single_byte_exist_in_vocab(self.byte_vocab)

        # inverted map of inv_byte_vocab (int index : bytes)
        self.inv_byte_vocab = {v: k for (k, v) in self.byte_vocab.items()}

        # encoded str : int index
        # convert bytes to our encoded form for keys
        self.str_vocab = {self.hfe.to_encoded(k): v for (k, v) in self.byte_vocab.items()}
        # int index : encoded str
        self.inv_str_vocab = {v: k for (k, v) in self.str_vocab.items()}

    @staticmethod
    def set_bytes_vocab(new_vocab):
        # bytes : int index
        byte_vocab = {}
        for idx, token in enumerate(new_vocab):
            # token should have been converted to bytes
            assert type(token) == bytes
            byte_vocab[token] = idx
        return byte_vocab

    def id_to_bytes(self, token_id):
        return self.inv_byte_vocab[token_id]

    def id_to_encoded(self, token_id):
        return self.inv_str_vocab[token_id]

    def get_vocabulary(self):
        return self.byte_vocab

    def vocab_size(self):
        return len(self.byte_vocab)

    # human readable for debugging
    def print_tokens(self, ids):
        return [self.inv_byte_vocab[i] for i in ids]

    # add all the single byte id's with a given score to some vocab
    def add_all_byte_ids(self, vocab, score=1e400):
        for i in range(256):
            # what is the corresponding token id
            tid = self.byte_vocab[bytes([i])]
            # add that with a "good" score
            vocab[tid] = score

    def tokenize(self, sent, tokens_only=False):
        if isinstance(sent, str):
            sent = bytes(sent, encoding='utf-8')
        data = []
        i = 0
        while i < len(sent):  # Iterate through the sentence input
            for j in range(self.max_len, 0, -1):  # Find the longest possible token
                tok = sent[i:i + j]
                if tok in self.byte_vocab:
                    if tokens_only:
                        # Only add token_id to results
                        data.append(self.byte_vocab[tok])
                    else:
                        # Add (token_id, token_start_idx, token_width) to results
                        data.append((self.byte_vocab[tok], i, len(tok)))
                    i += j  # advance to next token
                    break  # the for loop
        return data

    # return the tokenization as tokens in encoded str form
    def tokenize_to_encoded_str(self, sent):
        return [self.inv_str_vocab[token_id] for token_id in self.tokenize(sent, tokens_only=True)]

    # same but return byte form
    def tokenize_to_bytes(self, sent):
        return [self.inv_byte_vocab[token_id] for token_id in self.tokenize(sent, tokens_only=True)]

    # add the appropriate (t,v,v') triples to our dictionary
    # where t, v, and v' are all int indices
    @staticmethod
    def do_triples(combined, pad, padleft, padright, cur_id, sign, triples):
        # where the right padding starts
        right_ind = len(combined) - padright

        # iterate over the targets
        # note that the padding elements now have different contexts in
        # center section, so need to let them be targets too
        for t, target in enumerate(combined):
            # the contexts, need pad here not padleft or padright,
            # since some context may be within the combined
            for c in range(t - pad, t + pad + 1):
                # context is in range and distinct from target
                # ignore the case where both c and t are in padding since that cancels
                if c >= 0 and c != t and c < len(combined) and \
                        ((c >= padleft and c < right_ind) or (t >= padleft and t < right_ind)):
                    trip = (cur_id, target, combined[c])
                    # add sign to the triples
                    triples[trip] = triples.get(trip, 0) + sign

    # tokenize the sentence `sent`
    # add to the counts in the triples dict
    # tracking the (cur_id,t,c) for the ablated token cur_id,
    # with target token t and context token c
    # also updates the statistics in ablated_sizes
    # returns the total_tokens from tokenizing `sent`
    def fast_sage(self, sent, triples, ablated_sizes, pad=2, verbose=False):

        n = len(sent)

        # returns triples of (ids, start_index, width)
        values = self.tokenize(sent)
        ids, start_indices, widths = zip(*values)
        # if you use np.array here, remember to fix concatenation below

        # note, these are arrays over the tokens so len(values) < n
        total_tokens = len(values)

        # tuples to list
        ids = list(ids)
        start_indices = list(start_indices)
        widths = list(widths)

        max_len = 0

        # have a constant time lookup on whether we're at a token on the base tokenization
        # if >= 0, is the index of the token in ids or widths
        on_base = np.zeros(n, dtype=int) - 1
        for j, si in enumerate(start_indices):
            on_base[si] = j
        # now we can just produce our ablated tokenizations
        # quite efficiently
        for loc, (cur_id, start_index, width) in enumerate(values):
            # skip single bytes
            if width > 1:

                ablated_tokenization = []

                # find the next token with width-1 or less
                # starting at start_index
                i = start_index
                for j in range(width - 1, 0, -1):
                    tok = sent[i:i + j]
                    if tok in self.byte_vocab:
                        ablated_tokenization.append(self.byte_vocab[tok])  # keep the ids
                        i += j  # advance to next token
                        break  # the for loop

                # now extend as normal until we get back on the old path
                while i < n:
                    for j in range(min(self.max_len, n - i), 0, -1):
                        tok = sent[i:i + j]
                        if tok in self.byte_vocab:
                            ablated_tokenization.append(self.byte_vocab[tok])
                            i += j  # advance to next token
                            break  # the for loop

                    # we never got back on the path, so set beyond to n
                    if i >= n:
                        beyond = n
                        break

                    # we get to a spot on the current longest path
                    # we're back to the old tokenization, set beyond accordingly
                    if on_base[i] != -1:
                        beyond = on_base[i]
                        break

                if verbose:
                    print(self.print_tokens(ablated_tokenization))

                # track how many tokens were required for the ablation
                lat = len(ablated_tokenization)
                ablated_sizes[lat] = ablated_sizes.get(lat, 0) + 1
                max_len = max(max_len, lat)

                # note: on_base[i] is one beyond the last difference
                base_tok = ids[loc:beyond]
                if verbose:
                    print(self.print_tokens(base_tok))

                # can we do any padding on left or right
                padleft = min(pad, loc)
                padright = min(pad, len(values) - beyond)
                left_pad = ids[loc - padleft:loc]
                # print(print_tokens(left_pad))
                # note: beyond is one beyond the last difference
                right_pad = ids[beyond:beyond + padright]
                # print(print_tokens(right_pad))

                # combine with the padding, and work out the context triples
                combined_ab = left_pad + ablated_tokenization + right_pad
                self.do_triples(combined_ab, pad, padleft, padright, cur_id, 1, triples)

                # and same for the base tokenization
                combined_base = left_pad + base_tok + right_pad
                self.do_triples(combined_base, pad, padleft, padright, cur_id, -1, triples)

                if verbose:
                    print("base:", self.print_tokens(left_pad), self.print_tokens(base_tok),
                          self.print_tokens(right_pad))
                    print("ab:  ", self.print_tokens(left_pad), self.print_tokens(ablated_tokenization),
                          self.print_tokens(right_pad))
                    print("comb base:", self.print_tokens(combined_base))
                    print("comb ab:", self.print_tokens(combined_ab))
                    print()

        # log some of these
        if max_len > 200:
            # remember to convert from bytes
            print("long max_len:", max_len, '"' + sent.decode('utf-8') + '"')

        return total_tokens



# Copyright © 2023 Kensho Technologies, LLC

import argparse
import multiprocessing as mp
import random

from sage_tokenizer.SaGeVocabBuilder import SaGeVocabBuilder


def load_args():
    parser = argparse.ArgumentParser(description="Optimized implementation of SaGe method")
    parser.add_argument("experiment_name",
                        help="name of experiment, will save results under that name.")
    parser.add_argument("--corpus_filepath", required=True,
                        help="filepath for full corpus (e.g. wiki corpus)")
    parser.add_argument("--initial_vocabulary_filepath", required=True,
                        help="initial vocabulary, hex formatted, one vocab word per line")
    parser.add_argument("--vocabulary_schedule", nargs="+", type=int, required=True,
                        help="what vocabulary sizes are we aiming for. Tokenization won't be done on the last value")
    parser.add_argument("--embeddings_schedule", nargs="+", type=int, required=True,
                        help="from vocabulary_schedule, in which steps we should re-run embeddings")
    parser.add_argument("--partial_corpus_filepath", default="",
                        help="where to create / load partial corpus file. "
                             "Default is empty string for creating partial corpus under 'data' folder")
    parser.add_argument("--partial_corpus_line_number", type=int, default=1000,
                        help="number of lines for partial corpus - in thousands. Default is 1000")
    parser.add_argument("--max_len", type=int, default=16,
                        help="max length of tokens in bytes. Default is 16")
    parser.add_argument("--workers", type=int, default=max(1, mp.cpu_count() - 1),
                        help="number of worker threads to use. Default is max(1, mp.cpu_count()-1)")
    parser.add_argument("--random_seed", type=int, default=random.randint(1, 10000),
                        help="random seed value. Default is random.randint(1, 10000)")

    # word2vec params
    parser.add_argument("--word2vec_D", type=int, default=50,
                        help="word2vec embedding vector length. Default is 50")
    parser.add_argument("--word2vec_N", type=int, default=15,
                        help="word2vec number of negative samples. Default is 15")
    parser.add_argument("--word2vec_ALPHA", type=float, default=0.025,
                        help="word2vec Initial learning rate. Default is 0.025")
    parser.add_argument("--word2vec_window_size", type=int, default=5,
                        help="word2vec context window size. Default is 5")
    parser.add_argument("--word2vec_min_count", type=int, default=1,
                        help="word2vec minimum count of word. Default is 1, i.e. must be used at least once")
    parser.add_argument("--word2vec_sg", type=int, default=1,
                        help="word2vec skip-gram if 1; otherwise CBOW. Default is 1")

    return vars(parser.parse_args())


if __name__ == '__main__':
    args = load_args()
    vocab_builder = SaGeVocabBuilder(
        args['vocabulary_schedule'],
        args['embeddings_schedule'],
        args['max_len'],
        args['workers'],
        args['random_seed'],
        args['word2vec_D'],
        args['word2vec_N'],
        args['word2vec_ALPHA'],
        args['word2vec_window_size'],
        args['word2vec_min_count'],
        args['word2vec_sg']
    )

    vocab_builder.build_vocab(
        args['experiment_name'],
        args['corpus_filepath'],
        args['initial_vocabulary_filepath'],
        args['partial_corpus_filepath'],
        args['partial_corpus_line_number'],
    )



import numpy as np
import Parameters
import gensim.models

############################################################################

class CorpusIteratorForGensim:
    def __init__(self, corpus, model):
        self._corpus = corpus
        self._model = model

    def __iter__(self):
        corpus_lines = self._corpus.get_corpus()

        for line in corpus_lines:
            # yield tokenized line, when representing tokens as strings (not ints)
            tokens_in_line = [self._model.id_to_piece(x) for x in self._model.encode(line, out_type=int)]
            yield tokens_in_line

############################################################################

class EmbeddingsTrainer:
    def __init__(self, model, corpus, window_size, log):
        self._model = model
        self._corpus = corpus
        self._window_size = window_size
        self._log = log

    ######################################

    def construct_word2vec_embeddings(self, word2vec_model):
        vocab_size = self._model.vocab_size()
        embeddings = np.zeros(shape=(vocab_size, Parameters.D))
        
        for i in range(vocab_size):
            ith_token = self._model.id_to_piece(i)
            if ith_token in word2vec_model.wv.key_to_index.keys():    
                embeddings[i] = word2vec_model.wv[ith_token]
            else:
                self._log.warning("No wv for token {}. Assigning random vector...".format(ith_token))
                embeddings[i] = np.random.uniform(low=-0.5 / Parameters.D, high=0.5 / Parameters.D, size=(1, Parameters.D))

        return embeddings

    ######################################

    def train_embeddings(self):
        self._log.info("Training embeddings.")
        sentences = CorpusIteratorForGensim(self._corpus, self._model)
        self._log.info("Built CorpusIteratorForGensim")
        word2vec_model = gensim.models.Word2Vec(sentences=sentences, \
                                                vector_size=Parameters.D, \
                                                window=self._window_size,
                                                min_count=0,
                                                sg=1,
                                                negative=Parameters.N)
        self._log.info("gensim.models.Word2Vec Finished")

        # now extract the embeddings in the format we expect to
        embeddings = self.construct_word2vec_embeddings(word2vec_model)
        self._log.info("Finished construct_word2vec_embeddings")

        # in gensim there is no way to get context embeddings - so we do target=context=word vectors.
        return embeddings, embeddings

############################################################################



import Utils
import random

class Corpus:
    def __init__(self, corpus_filepath, partial_corpus_filepath, partial_corpus_lines_number, log):
        self._log = log

        with open(corpus_filepath) as full_corpus:
            self._full_corpus_data = full_corpus.readlines()
            self._log.info("original corpus num of lines: {}".format(len(self._full_corpus_data)))
            random.shuffle(self._full_corpus_data)
            self._partial_corpus_data = self._full_corpus_data[:partial_corpus_lines_number]
            self._log.info("num of lines in corpus: {}".format(len(self._partial_corpus_data)))

        with open(partial_corpus_filepath, "w+") as partial_corpus_file:
            partial_corpus_file.writelines(self._partial_corpus_data)

    def get_full_corpus(self):
        return self._full_corpus_data

    def get_partial_corpus(self):
        return self._partial_corpus_data

    def get_corpus(self, partial=True):
        if partial:
            cor = self.get_partial_corpus()
        else:
            cor = self.get_full_corpus()

        self._log.info("corpus num of lines: {}".format(len(cor)))
        return cor

    def compute_window(self, token_index, tokens_in_line):
        return Utils.compute_window(token_index, tokens_in_line)




# Copyright © 2023 Kensho Technologies, LLC

import json
import logging
import multiprocessing as mp
import os
import random
import time

import numpy as np
from scipy.special import expit


# only log code outside of multiprocessing
# logger = logging.getLogger(__name__)

# dump the vocab to a file, encoded as characters here
# no special tokens are added
# are saved in same order by index, so should preserve order
def write_vocab(vocab, filename):
    vocab_size = len(vocab)

    # write these in increasing index order
    # so same as any previous order
    byindex = sorted([(idx, token) for token, idx in vocab.items()])

    with open(filename, 'w', encoding="utf-8") as f:
        for _, token in byindex:
            f.write(token.hex() + '\n')


def save_sorted_losses(sage_model, sorted_losses, target_vocab_size, vocab_folder):
    sorted_losses_filepath = f"{vocab_folder}/sorted_losses_before_{target_vocab_size}.txt"
    logging.info(f"Saving sorted losses to {sorted_losses_filepath}")
    write_sorted_losses_into_file(sorted_losses, sorted_losses_filepath, sage_model)
    worst_500_filepath = f"{vocab_folder}/worst_500_{target_vocab_size}.txt"
    write_sorted_losses_into_file(sorted_losses[:500], worst_500_filepath, sage_model)
    best_500_filepath = f"{vocab_folder}/best_500_{target_vocab_size}.txt"
    write_sorted_losses_into_file(sorted_losses[-500:], best_500_filepath, sage_model)


def write_sorted_losses_into_file(sl, filename, sage_model):
    with open(filename, 'w', encoding="utf-8") as f:
        for loss, tid in sl:
            f.write(sage_model.id_to_encoded(tid) + "\t" + str(loss) + "\n")


# read our hex formatted vocab file
# return a list of bytes objects
# input file has one vocab word per line,
# each hex encoded
def load_vocab(vocab_filepath):
    if not os.path.exists(vocab_filepath):
        raise FileNotFoundError(f'Missing vocab file: {vocab_filepath}')

    with open(vocab_filepath) as vocab_file:
        # fromhex ignores whitespace from \n at end
        initial_vocab = [bytes.fromhex(token) for token in vocab_file.readlines()]

    return initial_vocab


def load_corpus(corpus_filepath, partial_corpus_filepath, partial_corpus_line_number):
    if os.path.exists(partial_corpus_filepath):
        # corpus already exists, directly loading
        logging.info(f"Found Processed Partial Corpus. Directly Loading from {partial_corpus_filepath}...")
        read_start = time.time()
        with open(partial_corpus_filepath) as corpus_f:
            partial_corpus = corpus_f.readlines()
        logging.info(f"Size of Corpus: {len(partial_corpus)}, time: {(time.time() - read_start):.2f}")
    else:
        read_start = time.time()
        with open(corpus_filepath) as full_corpus_f:
            corpus = full_corpus_f.readlines()
            logging.info(f"Loading from Original Corpus. Number of lines: {len(corpus)}")

        random.shuffle(corpus)
        logging.info(f"Original Corpus read and shuffled. Time: {(time.time() - read_start):.2f}")

        # may be same as original depending on partial_corpus_line_number
        write_start_time = time.time()
        partial_corpus = corpus[:partial_corpus_line_number * 1000]
        corpus_filename = os.path.splitext(os.path.basename(corpus_filepath))[0]

        if partial_corpus_filepath == "":
            if not os.path.exists("data"):
                logging.info("'data/' not found. Creating one for storing partial corpus...")
                os.mkdir("data")
            partial_corpus_filepath = f"data/{corpus_filename}_{len(partial_corpus)}.txt"

        with open(partial_corpus_filepath, "w+") as partial_corpus_f:
            partial_corpus_f.writelines(partial_corpus)
        logging.info(f"Partial corpus saved at {partial_corpus_filepath}. "
                     f"Number of lines: {len(partial_corpus)}, "
                     f"time: {(time.time() - write_start_time):.2f}")

    return partial_corpus


# Split the data given the number of chunks we expect
# Returns a generator
def divide_data_by_num(data, num_procs):
    size_per_chunk = len(data) // num_procs
    for i in range(0, len(data), size_per_chunk + 1):
        yield data[i: i + size_per_chunk + 1]


# Split the data given the size of chunks we expect
# Returns a generator
def divide_data_by_size(data, size):
    for i in range(0, len(data), size):
        yield data[i: i + size]


def verify_all_single_byte_exist_in_vocab(vocab):
    for i in range(256):
        b = bytes([i])
        if b not in vocab:
            raise Exception(f"missing byte {b}")


# function for computing losses given triple counts and embeddings
# losses : accumulate losses per ablated token, excluding the single byte ones, side effect this
# all_triples : triple values to aggregate into losses
# embeddings : embedding for each token
def compute_losses(losses, all_triples, embeddings):
    target_ids, context_ids, count = zip(*[(target_id, context_id, count) for (_, target_id, context_id), count in all_triples.items()])
    target_embeddings = np.array([embeddings[target_id] for target_id in target_ids])
    context_embeddings = np.array([embeddings[context_id] for context_id in context_ids])
    count = np.array(count)
    triples_loss = count * np.log(expit(np.einsum('ij,ij->i', target_embeddings, context_embeddings)))
    for idx, ((ablated_token_id, target_id, context_id), count) in enumerate(all_triples.items()):
        losses[ablated_token_id] = losses.get(ablated_token_id, 0.0) + triples_loss[idx]


def run_sage_parallel(embeddings, partial_corpus, sage_model, workers_number):
    logging.info(f"Splitting Data into {workers_number} chunks.")
    data_chunk_gen = divide_data_by_num(partial_corpus, workers_number)

    # these get aggregated over each chunk
    sage_losses = {}  # is token_id : loss
    overall_total_tokens = 0
    overall_total_triples = 0
    ablated_sizes = {}
    start_time = time.time()
    logging.info(f"Start spawning processes...")
    with mp.Pool(processes=workers_number) as pool:
        tasks = {}

        for tid, data_chunk in enumerate(data_chunk_gen):
            res = pool.apply_async(sage_per_chunk, args=(tid, sage_model, data_chunk, embeddings))
            tasks[res] = tid

        while tasks:
            results_ready_list = []
            for res, tid in tasks.items():
                if res.ready():
                    results_ready_list.append((res, tid))

            # process finished task results
            for res, tid in results_ready_list:
                losses, total_tokens, total_triples, ab_sizes = res.get()

                # just add these to totals/maxes
                overall_total_tokens += total_tokens
                overall_total_triples += total_triples

                # add to the overall tallys
                for k, v in losses.items():
                    sage_losses[k] = sage_losses.get(k, 0) + v

                # how many tokens needed to be examined
                for k, v in ab_sizes.items():
                    ablated_sizes[k] = ablated_sizes.get(k, 0) + v

                # all done with this,
                # can delete from tasks without messing up iteration over list
                del tasks[res]

                logging.info(f"task {tid} finished after {(time.time() - start_time):.2f} seconds. "
                             f"Tokens:{total_tokens}, triples:{total_triples}, active:{len(sage_losses)}")

            logging.info(f"Sleeping 1 second. Number of still running tasks: {len(tasks)}")
            time.sleep(1.0)
    return overall_total_tokens, overall_total_triples, sage_losses, ablated_sizes


# function that runs sage on each chunk of data (in parallelization)
# note: this is called from multiprocessing, so use print rather than logging
def sage_per_chunk(tid, model, data, embeddings, chunk_size=10000):
    print(f"Starting chunk {tid}, with {len(data)} lines of data")

    start_time = time.time()

    # accumulate over all the data
    losses = {}

    # these accumulate over each size
    triples = {}
    ablated_sizes = {}
    total_tokens = 0
    total_triples = 0
    total_fs_time = 0.0
    total_cl_time = 0.0

    fs_start = time.time()
    for row, d in enumerate(data):

        total_tokens += model.fast_sage(bytes(d, 'utf-8'), triples, ablated_sizes)

        # if filled up chunk, then compute the losses
        # to free up memory
        if (row > 0) and (row % chunk_size == 0):
            # take the total time here over all calls
            fs_time = time.time() - fs_start
            total_fs_time += fs_time
            # reinitialize fs_start
            fs_start = time.time()

            cl_start = time.time()
            compute_losses(losses, triples, embeddings)
            cl_time = time.time() - cl_start
            total_cl_time += cl_time

            print(f"fast_sage {tid}, row {row} of {len(data)}, "
                  f"fs_time: {fs_time:.2f}, cl_time: {cl_time:.2f}, "
                  f"triples: {len(triples)}, tokens: {total_tokens}")

            # total these up
            total_triples += len(triples)

            # zero out the triples from this chunksize lines
            triples = {}

    # compute for final partial chunk
    compute_losses(losses, triples, embeddings)
    total_triples += len(triples)

    # the triples can get quite large, so to avoid merging these
    # dict values, let's compute the losses in parallel too
    print(f"final fast_sage {tid}, row {row} of {len(data)}, "
          f"fs_time: {total_fs_time:.2f}, cl_time: {total_cl_time:.2f}, time: {(time.time() - start_time):.2f}, "
          f"triples: {len(triples)}, tokens: {total_tokens}")

    # Extra negative sign for equation (1) in SaGe paper
    # track number in cache too
    losses = {k: -v for k, v in losses.items()}

    return losses, total_tokens, total_triples, ablated_sizes


def init_logger(experiment_name):
    timestamp_str = time.strftime("%Y%m%d_%H%M%S")
    if not os.path.exists("logs"):
        os.mkdir("logs")
    log_filename = f'logs/{experiment_name}_{timestamp_str}.log'
    logging.basicConfig(filename=log_filename,
                        format="%(asctime)s - %(message)s",
                        datefmt="%m/%d/%Y %H:%M:%S",
                        level=logging.INFO)

    print(f"Logs will be stored in {log_filename}")


def get_output_folder(experiment_name):
    if not os.path.exists("results"):
        logging.info("\"results\" directory not found. Creating one for storing results")
        os.mkdir("results")
    results_path = "results/" + experiment_name
    if not os.path.exists(results_path):
        os.mkdir(results_path)
    vocab_folder = results_path + "/sage_vocabs"
    if not os.path.exists(vocab_folder):
        os.mkdir(vocab_folder)
    stats_folder = results_path + "/stats"
    if not os.path.exists(stats_folder):
        os.mkdir(stats_folder)
    embeddings_folder = results_path + "/embeddings"
    if not os.path.exists(embeddings_folder):
        os.mkdir(embeddings_folder)
    return embeddings_folder, stats_folder, vocab_folder


def set_random_seed(experiment_name, random_seed):
    seed_filepath = f"results/{experiment_name}/seed.txt"
    with open(seed_filepath, "w+") as f:
        f.write(str(random_seed))
    random.seed(random_seed)
    np.random.seed(random_seed)


def save_stats(stats, stats_folder, target_vocab_size):
    json_stats = json.dumps(stats, indent=2)  # pretty print a bit
    stats_filename = f"{stats_folder}/stats_{target_vocab_size}.json"
    logging.info(f"Saving stats to {stats_filename}")
    with open(stats_filename, 'wt') as f:
        f.write(json_stats + "\n")



# Copyright © 2023 Kensho Technologies, LLC

import os
import time
import logging

import gensim.models
import numpy as np


# class CorpusIterator:
#     def __init__(self, model, corpus_filepath):
#         self._model = model
#         self.corpus_filepath = corpus_filepath

#     def __iter__(self):
#         with open(self.corpus_filepath) as f:
#             corpus_data = f.readlines()
#         for line in corpus_data:
#             # convert bytes to tokens in encoded string form, for gensim
#             yield self._model.tokenize_to_encoded_str(bytes(line, 'utf-8'))

def get_embeddings(vocab_size, embeddings_folder, partial_corpus, sage_model, workers_number, word2vec_params):
    logging.info(f"training Embeddings at vocab size {vocab_size}")
    # is there an embedding of this size
    embeddings_filepath = f"{embeddings_folder}/embeddings_{vocab_size}.npy"
    if os.path.exists(embeddings_filepath):
        logging.info(f"Found trained embeddings. Loading it from {embeddings_filepath}")
        # context and target embeddings are the same so just keep one copy around
        embeddings = np.load(embeddings_filepath)
    else:
        logging.info(f"Start training embeddings with Word2Vec...")
        start_time = time.time()
        embeddings = train_embeddings(sage_model, partial_corpus, workers_number, word2vec_params)
        logging.info(f"Embeddings time:{time.time() - start_time}")
        logging.info(f"Save embeddings to {embeddings_filepath}")
        np.save(embeddings_filepath, embeddings, allow_pickle=True)
    return embeddings


def train_embeddings(sage_model, partial_corpus, workers, word2vec_params):
    # sentences = CorpusIterator(model, corpus_filepath)

    # also save in a version of this with a sentence per line, whitespace per token
    # which gensim's word2vec wants to process things in parallel
    # see https://github.com/RaRe-Technologies/gensim/releases/tag/3.6.0
    # and https://github.com/RaRe-Technologies/gensim/blob/develop/docs/notebooks/Any2Vec_Filebased.ipynb
    gensim_corpus_filepath = f"data/gensim_{sage_model.vocab_size()}.txt"

    # if already exists, use otherwise create the file
    if os.path.exists(gensim_corpus_filepath):
        logging.info(f"Gensim format data file already exists: {gensim_corpus_filepath}")
    else:
        gensim_start = time.time()
        logging.info(f"starting tokenization of {len(partial_corpus)} lines for gensim")
        with open(gensim_corpus_filepath, "w", encoding="utf-8") as gensim_file:
            for i, line in enumerate(partial_corpus):
                if i % 1000000 == 0:
                    logging.info(f"tokenizing line {i}, time: {(time.time() - gensim_start):.2f}")
                gensim_file.write(" ".join(sage_model.tokenize_to_encoded_str(bytes(line, 'utf-8'))) + "\n")
        logging.info(f"Gensim format data written: {gensim_corpus_filepath}, time: {(time.time()-gensim_start):.2f}")

    word2vec_model = gensim.models.Word2Vec(corpus_file=gensim_corpus_filepath,
                                            vector_size=word2vec_params.D,
                                            negative=word2vec_params.N,
                                            alpha=word2vec_params.ALPHA,
                                            window=word2vec_params.window_size,
                                            min_count=word2vec_params.min_count,
                                            sg=word2vec_params.sg,
                                            workers=workers)

    embeddings = np.zeros(shape=(sage_model.vocab_size(), word2vec_params.D))

    for idx, token in sage_model.inv_str_vocab.items():
        if token in word2vec_model.wv.key_to_index.keys():
            embeddings[idx] = word2vec_model.wv[token]
        else:
            # some may not have made the min_count value, so will be missing
            # Embeddings not found for this token. Assign a random vector
            # doing this the same way as the old SaGe code
            embeddings[idx] = np.random.uniform(low=-0.5/word2vec_params.D, high=0.5/word2vec_params.D, size=(1, word2vec_params.D))

    # just return one copy that we'll use for both context and target embeddings
    return embeddings



# Copyright © 2023 Kensho Technologies, LLC

# map all bytes to valid utf-8 characters
# in the same way that the huggingface tokenizers byte level pre-tokenizer does
class HFEncoding:

    # translated from rust code found here:
    # https://github.com/huggingface/tokenizers/blob/main/tokenizers/src/pre_tokenizers/byte_level.rs
    @staticmethod
    def bytes_char():

        bs = []
        bs.extend(range(ord('!'), ord('~') + 1))
        bs.extend(range(0xA1, 0xAC + 1))
        bs.extend(range(0xAE, 0xFF + 1))
        cs = [b for b in bs]

        n = 0
        for b in range(256):
            if b not in bs:
                bs.append(b)
                cs.append(2 ** 8 + n)
                n += 1

        return {bytes([f]): chr(t) for f, t in zip(bs, cs)}

    def __init__(self):
        # map any byte to the corresponding character
        self.byte_map = HFEncoding.bytes_char()
        # the inverse character to byte mapping
        self.inv_byte_map = {v: k for k, v in self.byte_map.items()}

    # convert an encoded string of our mapped characters back to the original bytes
    def to_bytes(self, s: str) -> bytes:
        return b"".join([self.inv_byte_map[c] for c in s])

    # convert a byte string into an encoded string of valid characters
    def to_encoded(self, byte_str: bytes) -> str:
        return "".join([self.byte_map[bytes([c])] for c in byte_str])
